﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Collections;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.ConditionActionEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Problems.ConditionActionClassification {
  [StorableClass]
  [Item("ConditionActionClassificationProblemData", "A problem data for LCS.")]
  public abstract class ConditionActionClassificationProblemData : ParameterizedNamedItem, IConditionActionProblemData {

    #region default data
    public static string[] defaultVariableNames = new string[] { "a", "b", "c", "d", "e", "f", "g" };
    public static double[,] defaultData = new double[,]{
      {0,0,1,1,0,0,0},
      {0,1,1,1,0,0,0},
      {0,0,1,0,0,0,1},
      {1,0,1,0,1,1,0}
    };
    #endregion

    #region parameter properites
    public IFixedValueParameter<Dataset> DatasetParameter {
      get { return (IFixedValueParameter<Dataset>)Parameters["Dataset"]; }
    }
    public IFixedValueParameter<ReadOnlyCheckedItemList<StringValue>> ConditionVariablesParameter {
      get { return (IFixedValueParameter<ReadOnlyCheckedItemList<StringValue>>)Parameters["ConditionVariables"]; }
    }
    public IFixedValueParameter<ReadOnlyCheckedItemList<StringValue>> ActionVariablesParameter {
      get { return (IFixedValueParameter<ReadOnlyCheckedItemList<StringValue>>)Parameters["ActionVariables"]; }
    }
    public IFixedValueParameter<IntRange> TrainingPartitionParameter {
      get { return (IFixedValueParameter<IntRange>)Parameters["TrainingPartition"]; }
    }
    public IFixedValueParameter<IntRange> TestPartitionParameter {
      get { return (IFixedValueParameter<IntRange>)Parameters["TestPartition"]; }
    }
    #endregion

    #region properties
    public Dataset Dataset {
      get { return DatasetParameter.Value; }
    }
    public ICheckedItemList<StringValue> ConditionVariables {
      get { return ConditionVariablesParameter.Value; }
    }
    public ICheckedItemList<StringValue> ActionVariables {
      get { return ActionVariablesParameter.Value; }
    }
    public IEnumerable<string> AllowedConditionVariables {
      get { return ConditionVariables.CheckedItems.Select(x => x.Value.Value); }
    }
    public IEnumerable<string> AllowedActionVariables {
      get { return ActionVariables.CheckedItems.Select(x => x.Value.Value); }
    }
    public IntRange TrainingPartition {
      get { return TrainingPartitionParameter.Value; }
    }
    public IntRange TestPartition {
      get { return TestPartitionParameter.Value; }
    }
    public IEnumerable<int> TrainingIndices {
      get {
        return Enumerable.Range(TrainingPartition.Start, Math.Max(0, TrainingPartition.End - TrainingPartition.Start))
                         .Where(IsTrainingSample);
      }
    }
    public IEnumerable<int> TestIndices {
      get {
        return Enumerable.Range(TestPartition.Start, Math.Max(0, TestPartition.End - TestPartition.Start))
           .Where(IsTestSample);
      }
    }
    public bool IsTrainingSample(int index) {
      return index >= 0 && index < Dataset.Rows &&
        TrainingPartition.Start <= index && index < TrainingPartition.End &&
        (index < TestPartition.Start || TestPartition.End <= index);
    }
    public bool IsTestSample(int index) {
      return index >= 0 && index < Dataset.Rows &&
             TestPartition.Start <= index && index < TestPartition.End;
    }

    public abstract IClassifierComparer ClassifierComparer { get; }
    #endregion

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      RegisterParameterEvents();
    }
    [StorableConstructor]
    protected ConditionActionClassificationProblemData(bool deserializing) : base(deserializing) { }
    protected ConditionActionClassificationProblemData(ConditionActionClassificationProblemData original, Cloner cloner)
      : base(original, cloner) {
      RegisterParameterEvents();
    }

    public ConditionActionClassificationProblemData(Dataset dataset, IEnumerable<string> allowedConditionVariables, IEnumerable<string> allowedActionVariables) {
      if (dataset == null) throw new ArgumentNullException("The dataset must not be null.");
      if (allowedActionVariables == null) throw new ArgumentNullException("The allowedActionVariables must not be null.");
      if (allowedConditionVariables == null) throw new ArgumentNullException("The allowedActionVariables must not be null.");

      var actionVariables = new CheckedItemList<StringValue>(dataset.VariableNames.Select(x => new StringValue(x)));
      var conditionVariables = new CheckedItemList<StringValue>(actionVariables);
      foreach (StringValue x in actionVariables) {
        actionVariables.SetItemCheckedState(x, allowedActionVariables.Contains(x.Value));
        conditionVariables.SetItemCheckedState(x, allowedConditionVariables.Contains(x.Value));
      }

      int trainingPartitionStart = 0;
      int trainingPartitionEnd = dataset.Rows / 2;
      int testPartitionStart = dataset.Rows / 2;
      int testPartitionEnd = dataset.Rows;

      Parameters.Add(new FixedValueParameter<Dataset>("Dataset", "", dataset));
      Parameters.Add(new FixedValueParameter<ReadOnlyCheckedItemList<StringValue>>("ActionVariables", "", actionVariables.AsReadOnly()));
      Parameters.Add(new FixedValueParameter<ReadOnlyCheckedItemList<StringValue>>("ConditionVariables", "", conditionVariables.AsReadOnly()));
      Parameters.Add(new FixedValueParameter<IntRange>("TrainingPartition", "", new IntRange(trainingPartitionStart, trainingPartitionEnd)));
      Parameters.Add(new FixedValueParameter<IntRange>("TestPartition", "", new IntRange(testPartitionStart, testPartitionEnd)));

      ((ValueParameter<Dataset>)DatasetParameter).ReactOnValueToStringChangedAndValueItemImageChanged = false;

      RegisterParameterEvents();
    }

    public event EventHandler Changed;
    protected virtual void OnChanged() {
      var listeners = Changed;
      if (listeners != null) listeners(this, EventArgs.Empty);
    }

    public IEnumerable<IInput> FetchInput(IEnumerable<int> rows) {
      foreach (var row in rows) {
        yield return FetchInput(row);
      }
    }
    public abstract IInput FetchInput(int rowNumber);

    public IEnumerable<IAction> FetchAction(IEnumerable<int> rows) {
      foreach (var row in rows) {
        yield return FetchAction(row);
      }
    }
    public abstract IAction FetchAction(int rowNumber);

    protected IDictionary<int, IInput> fetchInputCache = new Dictionary<int, IInput>();

    #region events
    private void RegisterParameterEvents() {
      ConditionVariablesParameter.ValueChanged += new EventHandler(VariablesChanged);
      ConditionVariablesParameter.Value.CheckedItemsChanged += new CollectionItemsChangedEventHandler<IndexedItem<StringValue>>(VariablesChanged);
      ActionVariablesParameter.ValueChanged += new EventHandler(VariablesChanged);
      ActionVariablesParameter.Value.CheckedItemsChanged += new CollectionItemsChangedEventHandler<IndexedItem<StringValue>>(VariablesChanged);
    }
    private void DeregisterParameterEvents() {
      ActionVariablesParameter.Value.CheckedItemsChanged += new CollectionItemsChangedEventHandler<IndexedItem<StringValue>>(VariablesChanged);
      ActionVariablesParameter.ValueChanged += new EventHandler(VariablesChanged);
      ConditionVariablesParameter.Value.CheckedItemsChanged += new CollectionItemsChangedEventHandler<IndexedItem<StringValue>>(VariablesChanged);
      ConditionVariablesParameter.ValueChanged += new EventHandler(VariablesChanged);
    }
    private void Value_CheckedItemsChanged(object sender, CollectionItemsChangedEventArgs<IndexedItem<StringValue>> e) {
      ActionConditionVariablesChanged();
    }
    private void VariablesChanged(object sender, EventArgs e) {
      ActionConditionVariablesChanged();
    }

    protected abstract void ActionConditionVariablesChanged();
    #endregion

    #region IDataAnalysisProblemData Members
    public bool IsEmpty {
      get { return true; }
    }
    public ICheckedItemList<StringValue> InputVariables {
      get { return ConditionVariables; }
    }
    public IEnumerable<string> AllowedInputVariables {
      get { return AllowedConditionVariables; }
    }
    #endregion
  }
}
